/*************************************************************************
 *                                                                          
 * qscanplotdoc.cpp  -  description
 *
 * Copyright (C) 2001  J C Gonzalez
 * gonzalez@gae.ucm.es
 *
 *------------------------------------------------------------------------
 *
 * Copyright (C) 2001 J C Gonzalez
 *  
 * This program is free software;  you can redistribute it and/or  modify
 * it under the terms  of the GNU General  Public License as published by
 * the Free Software Foundation; either version  2 of the License, or (at
 * your option) any later version.
 * 
 * This piece of code is distributed in the hope  that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of FITNESS
 * FOR A PARTICULAR PURPOSE.
 * 
 * In no  event shall his author  be liable for  any special, incidental,
 * indirect  or  consequential  damages  of any  kind,   or  any  damages
 * whatsoever resulting from loss of use, data or profits, whether or not
 * advised of the possibility of damage, and on  any theory of liability,
 * arising out  of or in connection  with the use  or performance of this
 * software. You've been warned.
 * 
 ************************************************************************/


#include "qscanplotdoc.h"

#include <qfile.h>
#include <qmsgbox.h>

#include "resource.h"
#include "math.h"

extern QMap<int, QString> Plugins_List;
extern int numPlugins;

QScanPlotDoc::QScanPlotDoc()
{
  modified = false;
  newDoc();
}

QScanPlotDoc::~QScanPlotDoc()
{
}

void QScanPlotDoc::newDoc()
{
  if (modified) {
    int loose=QMessageBox::warning(0, tr("Creating New Document..."),
          tr("Do you want to loose the changes made to your current document?"),
          QMessageBox::Yes, QMessageBox::No);

    if (loose==QMessageBox::No)
      return;
  }

  image.reset();
  imageName = "";
  theFileName = "-";

  int i;

  if (theCurves) {
    for (i=0; i<nCurves; i++)
      theCurves->remove(i);
    delete theCurves;
  }
  theCurves = new QIntDict<Curve>(31);
  theCurves->setAutoDelete(TRUE);
  nCurves = 0;

  if (theAxes) {
    for (i=0; i<nAxes; i++)
      theAxes->remove(i);
    delete theAxes;
  }
  theAxes   = new QIntDict<Axes>(11);
  theAxes->setAutoDelete(TRUE);
  nAxes = 0;

  if (theGrids) {
    for (i=0; i<nGrids; i++)
      theGrids->remove(i);
    delete theGrids;
  }
  theGrids   = new QIntDict<Grid>(11);
  theGrids->setAutoDelete(TRUE);
  nGrids = 0;

  emit documentChanged();

  modified = false;
}

bool QScanPlotDoc::save()
{
  return saveAs(theFileName);
}

bool QScanPlotDoc::saveAs(const QString &filename)
{
  QFile f( filename );
  f.open( IO_WriteOnly );                     // open file for writing
  QTextStream s( &f );                        // serialize using f

  s << *this;

  f.close();

  theFileName = filename;

  modified = false;

  return projectData(0,0);
}

bool QScanPlotDoc::load(const QString &filename)
{
  // parse xml file
  QspParser handler;
  handler.setDoc(this);
  QFile xmlFile( filename );
  QXmlInputSource source( xmlFile );
  QXmlSimpleReader reader;
  reader.setContentHandler( &handler );
  reader.setErrorHandler(&handler);
  bool ok = reader.parse( source );
  xmlFile.close();

  if (!ok) {
    QMessageBox::critical (0, tr("Parse Error"),
                              tr(handler.errorProtocol()));
    return false;
  }

  theFileName = filename;

  if (! loadImage( imageName ))
    return false;

  modified = false;

  emit documentChanged();
  return true;
}

QTextStream& operator<<(QTextStream & output, const QScanPlotDoc & aDoc)
{
  int i;

  //------------------------------------------------------------
  
  output << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
         << "<!DOCTYPE qscanplot>\n\n";

  output << "<qscanplot>\n\n";

  output << "  <image src=\"" << aDoc.imageName << "\"/>\n\n";

  //------------------------------------------------------------
  
  output << "  <axessystems number=\"" << aDoc.nAxes << "\">\n\n";
  
  Axes *a;
  for (i=0; i<aDoc.nAxes; i++) {
    a = aDoc.theAxes->find(i);
    output << *a;
  }

  output << "  </axessystems>\n\n";

  //------------------------------------------------------------
  
  output << "  <grids number=\"" << aDoc.nGrids << "\">\n\n";

  Grid *g;
  for (i=0; i<aDoc.nGrids; i++) {
    g = aDoc.theGrids->find(i);
    output << *g;
  }

  output << "  </grids>\n\n";

  //------------------------------------------------------------

  output << "  <plugins number=\"" << numPlugins << "\">\n";

  for (i=0; i<numPlugins; i++) {
    output << "    <plugin index=\"" << i
           << "\" name=\"" << Plugins_List[i] << "\"/>\n";
  }

  output << "  </plugins>\n\n";

  //------------------------------------------------------------

  output << "  <curves number=\"" << aDoc.nCurves << "\">\n\n";

  Curve *c;
  for (i=0; i<aDoc.nCurves; i++) {
    c = aDoc.theCurves->find(i);
    output << *c;
  }

  output << "  </curves>\n\n";

  //------------------------------------------------------------
  
  output << "</qscanplot>\n";

  //------------------------------------------------------------
  
  return output;
}

bool QScanPlotDoc::isModified() const
{
  return modified;
}

void QScanPlotDoc::setModified(bool m)
{
  modified = m;
}

/** Loads image file in any of the supported formats */
bool QScanPlotDoc::loadImage(const char* filename)
{
  bool ok = true;

  if ( filename )
    ok = image.load(filename, 0);

  imageName = QString(filename);

  modified = true;

	return ok;
}

/** Exports image */
QImage* QScanPlotDoc::getImage()
{
	return &image;
}

/** Returns pointer to i-th curve */
Curve* QScanPlotDoc::getCurve(int i) const
{
  return theCurves->find(i);
}

/** Returns pointer to i-th axes */
Axes* QScanPlotDoc::getAxes(int i) const
{
  return theAxes->find(i);
}

/** Returns pointer to i-th grid */
Grid* QScanPlotDoc::getGrid(int i) const
{
  return theGrids->find(i);
}

/** Returns pointer to i-th curve */
Curve* QScanPlotDoc::getCurrentCurve() const
{
  return theCurves->find(currentCurve);
}

/** Returns pointer to i-th axes */
Axes* QScanPlotDoc::getCurrentAxes() const
{
  return theAxes->find(currentAxes);
}

/** Returns number of Curves defined in current document */
int QScanPlotDoc::getNCurves() const
{
  return nCurves;
}

/** Returns number of Axes systems defined in current document */
int QScanPlotDoc::getNAxes() const
{
  return nAxes;
}

/** Returns number of Grids defined in current document */
int QScanPlotDoc::getNGrids() const
{
  return nGrids;
}

/** Selects curve c */
void QScanPlotDoc::selectCurve(int c)
{
  currentCurve = c;
}

/** Selects Axes system */
void QScanPlotDoc::selectAxes(int a)
{
  currentAxes = a;
}

/** Adds new Curve to list */
void QScanPlotDoc::newCurve(Curve* c)
{
  c->setCanvas(theCanvas);
	theCurves->insert(nCurves, c);
	nCurves++;
	modified = true;
}

/** Adds new Axes system to list */
void QScanPlotDoc::newAxes(Axes* a)
{
  a->setCanvas(theCanvas);
	theAxes->insert(nAxes, a);
	nAxes++;
	modified = true;
}

/** Adds new Grid to list */
void QScanPlotDoc::newGrid(Grid* g)
{
	theGrids->insert(nGrids, g);
	nGrids++;
	modified = true;
}

/** Removes Curve from list */
void QScanPlotDoc::deleteCurve(int n)
{
	theCurves->remove( n );
	nCurves--;
	selectCurve( 0 );
	modified = true;
	emit( deletedCurve( n ) );
}

/** Removes Axes from list */
void QScanPlotDoc::deleteAxes(int n)
{
	theAxes->remove( n );
	nAxes--;
	selectAxes( 0 );
	modified = true;
	emit( deletedAxes( n ) );
}

/** Removes Grid from list */
void QScanPlotDoc::deleteGrid(int n)
{
	theGrids->remove( n );
	nGrids--;
	modified = true;
}

/** Changes old Axes dict. with new one, and deletes the old */
void QScanPlotDoc::useBackupAxes(QIntDict<Axes> * dict)
{
  Axes * a;
  int i;
  for (i=0; i<nAxes; i++) {
    a = theAxes->find(i);
  	delete a;
  }
  delete theAxes; 	
  theAxes = dict;
}

/** Changes old Curves dict. with new one, and deletes the old */
void QScanPlotDoc::useBackupCurves(QIntDict<Curve> * dict)
{
  Curve * c;
  int i;
  for (i=0; i<nCurves; i++) {
    c = theCurves->find(i);
  	delete c;
  }
  delete theCurves; 	
  theCurves = dict;
}

/*
inline double doScale(int n, double x) {
  if (n == 0)
    return x;
  else
    return log10(x);
}

inline double undoScale(int n, double x) {
  if (n == 0)
    return x;
  else
    return pow(10.0, x);
}
*/

/** Projects data of Curve nC using Axes system nA */
bool QScanPlotDoc::projectData(int nA, int nC)
{
  static double ax, ay, bx, by, cx, cy;
  static double Ax, Ay, Bx, By, Cx, Cy;
  static double ux, uy, vx, vy;
  static double Ux, Uy, Vx, Vy;
  static double px, py, qx, qy;
  static double D, dl, dm, l, m;
  static double Px, Py;

  static Axes* a;
  static int sX;
  static int sY;

  if ( (nA >= nAxes) || (nC >= nCurves) )
    return false;

  a = getAxes(nA);

  if ( a->getNPoints() < 3 )
    return false;

  ax = a->getPoint(0)->getGX();
  bx = a->getPoint(1)->getGX();
  cx = a->getPoint(2)->getGX();

  ay = a->getPoint(0)->getGY();
  by = a->getPoint(1)->getGY();
  cy = a->getPoint(2)->getGY();

  ux = bx - ax,  uy = by - ay;
  vx = cx - ax,  vy = cy - ay;

  sX = a->getXAxisScale();
  sY = a->getYAxisScale();

  Ax = a->getPoint(0)->getX();
  Bx = a->getPoint(1)->getX();
  Cx = a->getPoint(2)->getX();

  Ay = a->getPoint(0)->getY();
  By = a->getPoint(1)->getY();
  Cy = a->getPoint(2)->getY();

  doScale(&Ax, &Ay, sX, sY);
  doScale(&Bx, &By, sX, sY);
  doScale(&Cx, &Cy, sX, sY);

  Ux = Bx - Ax,  Uy = By - Ay;
  Vx = Cx - Ax,  Vy = Cy - Ay;

  D  = (ux * vy) - (vx * uy);

  CurvePoint *p;
  PointListIterator it(*(getCurve(nC)->getPointList())); // iterator for list

  for ( ; it.current(); ++it ) {

    p = it.current();

    px = p->getGX();
    py = p->getGY();

    qx = px - ax,  qy = py - ay;

    dl = (qx * vy) - (vx * qy);
    dm = (ux * qy) - (qx * uy);

    l = dl / D;
    m = dm / D;

    Px = (l * Ux) + (m * Vx) + Ax;
    Py = (l * Uy) + (m * Vy) + Ay;

    undoScale(&Px, &Py, sX, sY);

    p->setX(Px);
    p->setY(Py);

  }

  return true;
}
